import pyparsing as pp

from .parser.grammar import module
from .parser.indent import reset_level


def get_exception_message(e: pp.ParseSyntaxException) -> str:
    return f"{e.lineno}:{e.column} {e.msg}, found {e.found}"


pp.ParseSyntaxException.__str__ = get_exception_message  # type: ignore


def check_format(content: str) -> list[pp.ParseBaseException]:
    reset_level()
    errors: list[pp.ParseBaseException] = []
    try:
        module.parseString(content, parseAll=True)
    except pp.ParseSyntaxException as e:
        errors.append(e)
    return errors



import pytest
from pyparsing import ParseException

from rapidchecker.parser.identifiers import elem_index, identifier, parameter, variable


@pytest.mark.parametrize("valid_identifier", ["VARIABLE", "test_snake", "testCamel"])
def test_valid_identifier(valid_identifier: str) -> None:
    result = identifier.parseString(valid_identifier, parseAll=True).as_list()
    assert result == [valid_identifier]


@pytest.mark.parametrize("invalid_identifier", ["MODULE", "1dsads", "-dsadsa"])
def test_invalid_identifier(invalid_identifier: str) -> None:
    with pytest.raises(ParseException):
        identifier.parseString(invalid_identifier, parseAll=True)


@pytest.mark.parametrize(
    ("index", "expected"),
    [("*", "*"), ("111", 111), ("variable_name", "variable_name")],
)
def test_elem_index(index: str, expected: str | int) -> None:
    result = elem_index.parseString(f"{{{index}}}", parseAll=True).as_list()
    assert result == ["{", expected, "}"]


@pytest.mark.parametrize(
    "valid_variable",
    [
        "VARIABLE",
        "test_snake",
        "testCamel",
        "with_index{1}",
        "object.attribute{index}",
        "objects{1}.attribute",
    ],
)
def test_valid_variable(valid_variable: str) -> None:
    assert variable.parseString(valid_variable, parseAll=True).as_list()


@pytest.mark.parametrize("invalid_variable", ["MODULE,dsa", "1dsads", "{1}dsadsa"])
def test_invalid_variable(invalid_variable: str) -> None:
    with pytest.raises(ParseException):
        variable.parseString(invalid_variable, parseAll=True)


@pytest.mark.parametrize(
    "valid_parameter",
    [
        "INOUT string name{100}",
        "\\INOUT string name{100}",
        "string name",
        "\\robtarget target",
        "VAR signaldo signal",
    ],
)
def test_parameter(valid_parameter: str) -> None:
    assert parameter.parseString(valid_parameter, parseAll=True)



import pyparsing as pp

from . import tokens as T
from .literals import NUMBER

identifier = ~T.RESERVED_WORD + pp.pyparsing_common.identifier
identifier.set_name("identifier")

datatype = identifier.set_name("datatype")

elem_index = pp.Optional("{" + (identifier | NUMBER | "*") + "}").set_name("index")

variable = identifier + elem_index + pp.ZeroOrMore("." - identifier + elem_index)
variable.set_name("variable")


parameter = (
    pp.Optional("\\")
    + pp.Optional(T.INOUT | T.PERS | T.VAR)
    + datatype
    + identifier
    + elem_index
)
parameter.set_name("param")
parameter_list = pp.Optional(pp.delimitedList(parameter))



from pathlib import Path

from rapidchecker.whitespace_checks import (
    NoTrailingLineError,
    TooManyEmptyLinesError,
    TrailingSpaceError,
    check_whitespace,
)


def test_check_whitespace_no_errors() -> None:
    file_contents = Path("./tests/rapid_samples/SimpleModule.sys").read_text()
    result = check_whitespace(file_contents)
    assert not result


def test_check_whitespace_with_errors() -> None:
    file_contents = Path("./tests/rapid_samples/BadWhitespaceModule.sys").read_text()
    result = check_whitespace(file_contents)
    expected = [
        TrailingSpaceError(
            lineno=4,
            col=26,
            error="Line ends with space",
            suggestion="Remove trailing spaces",
        ),
        TooManyEmptyLinesError(
            lineno=7,
            col=0,
            error="Too many empty lines",
            suggestion="Use up to 2 empty lines",
        ),
        TooManyEmptyLinesError(
            lineno=8,
            col=0,
            error="Too many empty lines",
            suggestion="Use up to 2 empty lines",
        ),
        NoTrailingLineError(
            lineno=27,
            col=0,
            error="Line does not end with a new line character",
            suggestion="Add a new line character at the end of the line",
        ),
    ]

    assert result == expected



import pyparsing as pp

from rapidchecker.config import CONFIG

from . import tokens as T
from .identifiers import datatype, identifier, parameter_list, variable
from .indent import (
    INDENT,
    INDENT_CHECKPOINT,
    UNDENT,
    register_indent_checks,
)
from .literals import RAPIDLITERAL
from .operators import infix_op, prefix_op

pp.ParserElement.enable_packrat()

# Simple rules for entire lines
comment = "!" + pp.rest_of_line
eval_stmt = "%" - pp.rest_of_line  # TODO: Fix this to account for the last %

# These elements are needed to define the next rules
expression = pp.Forward()
stmt_block = pp.Forward()
stmt = pp.Forward()

# Arguments for proc and functon calls
optional_arg = pp.Combine("\\" + identifier) + pp.Optional(":=" + expression)
required_arg = pp.Optional(identifier + ":=") + expression
first_argument = optional_arg | required_arg
argument = ("," + (optional_arg | required_arg)) | optional_arg
argument_list = pp.Optional(first_argument + pp.ZeroOrMore(argument))

function_call = identifier + "(" - argument_list - ")"

array = "[" + pp.delimitedList(expression) + "]"

# Terms and expressions can be defined now
term = (
    function_call
    | variable
    | RAPIDLITERAL
    | "(" + expression + ")"
    | prefix_op + expression
    | array
)
expression <<= (term + infix_op + expression) | term
assignment = variable + ":=" - expression - ";"

# Record definition
record_item = datatype + identifier + ";"
record_def = (
    T.RECORD
    - identifier
    - INDENT
    - pp.OneOrMore(record_item)
    - UNDENT
    - INDENT_CHECKPOINT
    - T.ENDRECORD
)
record_def_section = pp.ZeroOrMore(record_def)

# Variable definition
var_def = (
    pp.Optional(T.LOCAL | T.TASK)
    + (T.PERS | T.VAR | T.CONST)
    + datatype
    - identifier
    - pp.Optional("{" + expression + "}")
    - pp.Optional(":=" + expression)
    - ";"
)
var_def_section = pp.ZeroOrMore(var_def)

# If statement
if_stmt = (
    T.IF
    + expression
    + T.THEN
    - stmt_block
    - pp.ZeroOrMore(INDENT_CHECKPOINT + T.ELSEIF - expression - T.THEN - stmt_block)
    - pp.Optional(INDENT_CHECKPOINT + T.ELSE - stmt_block)
    - INDENT_CHECKPOINT
    - T.ENDIF
)

# Test statement
test_stmt = (
    T.TEST
    - expression
    - pp.OneOrMore(INDENT_CHECKPOINT + T.CASE - term - ":" - stmt_block)
    - pp.Optional(INDENT_CHECKPOINT + T.DEFAULT - ":" - stmt_block)
    - INDENT_CHECKPOINT
    - T.ENDTEST
)

# Loops
while_stmt = T.WHILE - expression - T.DO - stmt_block - T.ENDWHILE
for_stmt = (
    T.FOR
    - identifier
    - T.FROM
    - expression
    - T.TO
    - expression
    - pp.Optional(T.STEP + expression)
    - T.DO
    - stmt_block
    - INDENT_CHECKPOINT
    - T.ENDFOR
)

# Simple statements
proc_call_stmt = identifier - argument_list - ";"
func_call_stmt = function_call - ";"
connect_stmt = T.CONNECT - identifier - T.WITH - identifier - ";"
return_stmt = T.RETURN - pp.Optional(expression) - ";"

compact_if_stmt = (
    T.IF
    + expression
    + (assignment | func_call_stmt | proc_call_stmt | return_stmt | connect_stmt)
)

stmt <<= (
    assignment
    | if_stmt
    | compact_if_stmt
    | test_stmt
    | while_stmt
    | for_stmt
    | eval_stmt
    | func_call_stmt
    | proc_call_stmt
    | return_stmt
    | connect_stmt
)
stmt_block <<= INDENT - var_def_section - pp.ZeroOrMore(stmt) - UNDENT

# Function definition
func_def = (
    T.FUNC
    - datatype
    - identifier
    - "("
    - parameter_list
    - ")"
    - stmt_block
    - INDENT_CHECKPOINT
    - T.ENDFUNC
)

# Proc definition
if CONFIG.indent_error_section:
    error_section = T.ERROR - INDENT - stmt_block - UNDENT
else:
    error_section = INDENT_CHECKPOINT + T.ERROR - stmt_block
proc_def = (
    T.PROC
    - identifier
    - "("
    - parameter_list
    - ")"
    - stmt_block
    - pp.Optional(error_section)
    - INDENT_CHECKPOINT
    - T.ENDPROC
)

# Trap definition
trap_def = T.TRAP - identifier - stmt_block - INDENT_CHECKPOINT - T.ENDTRAP

def_section = pp.ZeroOrMore(func_def | proc_def | trap_def)

module_args = "(" - pp.delimitedList(T.MODULE_OPTIONS) - ")"
module = (
    T.MODULE
    - identifier
    - pp.Optional(module_args)
    - INDENT
    - record_def_section
    - var_def_section
    - def_section
    - UNDENT
    - INDENT_CHECKPOINT
    - T.ENDMODULE
)
module.ignore(comment)

register_indent_checks(
    [
        stmt,
        var_def,
        record_item,
        T.PROC,
        T.MODULE,
        T.FUNC,
        T.TRAP,
        T.RECORD,
        T.IF,
        T.FOR,
        T.WHILE,
        T.TEST,
    ],
)



import pytest
from pyparsing import ParseException

from rapidchecker.parser.tokens import MODULE, RESERVED_WORD


def test_rapid_keyword() -> None:
    assert MODULE.parseString("MODULE", parseAll=True).as_list() == ["MODULE"]


def test_lowercase_rapid_keyword() -> None:
    with pytest.raises(ParseException):
        assert MODULE.parseString("module", parseAll=True)


def test_reserved_word() -> None:
    assert RESERVED_WORD.parseString("MODULE", parseAll=True).as_list() == ["MODULE"]



import pytest
from pyparsing import ParseException, ParseSyntaxException

from rapidchecker.parser.grammar import (
    argument_list,
    array,
    assignment,
    comment,
    compact_if_stmt,
    connect_stmt,
    def_section,
    eval_stmt,
    expression,
    for_stmt,
    func_call_stmt,
    func_def,
    function_call,
    if_stmt,
    module,
    optional_arg,
    proc_call_stmt,
    proc_def,
    record_def,
    return_stmt,
    stmt,
    stmt_block,
    term,
    test_stmt,
    trap_def,
    var_def,
    var_def_section,
    while_stmt,
)
from rapidchecker.parser.indent import reset_level


def test_comment() -> None:
    result = comment.parseString("!This is a comment", parseAll=True).as_list()
    assert result == ["!", "This is a comment"]


def test_eval() -> None:
    result = eval_stmt.parseString("%procCall%", parseAll=True).as_list()
    assert result == ["%", "procCall%"]


@pytest.mark.parametrize(
    "valid_expression",
    [
        "a + 1",
        "a{100}",
        "b.c.d{*} AND TRUE",
        "-a",
        "-10",
        "-funCall(a,b)",
        "a = b AND b <> z",
    ],
)
def test_expression(valid_expression: str) -> None:
    assert expression.parseString(valid_expression, parseAll=True).as_list()


@pytest.mark.parametrize(
    "input_str",
    ["a := 1;", "RETURN TRUE;", "procCall;", 'a := funcCall(1, "string");'],
)
def test_statement(input_str: str) -> None:
    assert stmt.parseString(input_str, parseAll=True).as_list()


@pytest.mark.parametrize(
    "input_str",
    ["a == 1;", "RETURN TRUE"],
)
def test_invalid_statement(input_str: str) -> None:
    with pytest.raises(ParseSyntaxException):
        stmt.parseString(input_str, parseAll=True)


@pytest.mark.parametrize(
    "valid_block",
    ["  a := 1;\n  b:=c;", "  RETURN TRUE;\n  a:=c;"],
)
def test_stmt_block(valid_block: str) -> None:
    reset_level()
    assert stmt_block.parseString(valid_block, parseAll=True).as_list()


@pytest.mark.parametrize("input_str", ["\\a", "\\a:=c"])
def test_optional_arg(input_str: str) -> None:
    assert optional_arg.parseString(input_str, parseAll=True).as_list()


@pytest.mark.parametrize("input_str", ["\\a", "\\a, \\b:=c", "\\a, c+1", "a, \\b:=c"])
def test_arg_list(input_str: str) -> None:
    assert argument_list.parseString(input_str, parseAll=True).as_list()


def test_empty_arg_list() -> None:
    assert argument_list.parseString("", parseAll=True).as_list() == []


@pytest.mark.parametrize("input_str", ["a(\\a)", "funcName()", "funcName(c+1, \\a)"])
def test_function_call(input_str: str) -> None:
    assert function_call.parseString(input_str, parseAll=True).as_list()


@pytest.mark.parametrize(
    "input_str",
    ["[TRUE AND FALSE]", "[1,2]", '[b + c, "dsads", "dsadsa"]'],
)
def test_array(input_str: str) -> None:
    assert array.parseString(input_str, parseAll=True).as_list()


@pytest.mark.parametrize(
    "input_str",
    [
        "funcCall(a,\\a:=1)",
        "varName",
        "TRUE",
        '"string"',
        "(1+1)",
        "NOT a AND B",
        "-(a AND B)",
        "[1,2,3,4]",
    ],
)
def test_term(input_str: str) -> None:
    assert term.parseString(input_str, parseAll=True).as_list()


@pytest.mark.parametrize(
    "input_str",
    ["a:=b;", "a:=(1+2+3);", "object.attr := [1,2,34];", "objects{1}.attr := NOT b;"],
)
def test_assignment(input_str: str) -> None:
    assert assignment.parseString(input_str, parseAll=True).as_list()


@pytest.mark.parametrize(
    "input_str",
    ["RECORD test\n  string a;\n  robtarget target;\n  num a;\nENDRECORD"],
)
def test_record_def(input_str: str) -> None:
    reset_level()
    assert record_def.parseString(input_str, parseAll=True).as_list()


@pytest.mark.parametrize(
    "input_str",
    [
        "PERS string varName;",
        "VAR robtarget targets{1000};",
        "VAR robtarget targets{var1 + var2};",
        "CONST num number := 1 + 1;",
        "LOCAL CONST num number := 1 + 1;",
    ],
)
def test_var_def(input_str: str) -> None:
    assert var_def.parseString(input_str, parseAll=True).as_list()


@pytest.mark.parametrize(
    "input_str",
    [
        "PERS string varName;\nVAR robtarget targets{1000};\nCONST num number := 1 + 1;",
    ],
)
def test_var_def_section(input_str: str) -> None:
    assert var_def_section.parseString(input_str, parseAll=True).as_list()


@pytest.mark.parametrize(
    "input_str",
    [
        "IF a THEN\n  callProc;\nELSE\n  callAnotherProc;\nENDIF",
        "IF a THEN\n  callProc;\nELSEIF new_condition AND B THEN\n  callNewProc;\nELSE\n  callAnotherProc;\nENDIF",
        "IF a THEN\n  callProc;\n  callProc2;\nELSEIF new_condition AND B THEN\n  callNewProc;\n  callProc2;\nELSE\n  callAnotherProc;\nENDIF",
        "IF a = b AND b <> 1 THEN\n  var1 := 0;\nENDIF",
    ],
)
def test_valid_if_stmt(input_str: str) -> None:
    reset_level()
    assert if_stmt.parseString(input_str, parseAll=True).as_list()


@pytest.mark.parametrize(
    "input_str",
    [
        "IF a THEN\n  callProc;\nELSE\n  callAnotherProc;",
        "IF a callProc;",
        "IF a THEN\n  callProc;\nELSEIF new_condition AND B\n  callNewProc;\nELSE\n  callAnotherProc;\nENDIF",
    ],
)
def test_invalid_if_stmt(input_str: str) -> None:
    reset_level()
    with pytest.raises((ParseException, ParseSyntaxException)):
        if_stmt.parseString(input_str, parseAll=True)


# TODO: Why does this fail?
@pytest.mark.parametrize(
    "input_str",
    [
        "IF a callProc;",
    ],
)
def test_compact_if(input_str: str) -> None:
    assert compact_if_stmt.parseString(input_str, parseAll=True).as_list()


@pytest.mark.parametrize(
    "input_str",
    [
        "TEST a\nCASE 1:\n  callProc1;\n  callProc2;\nCASE 2:\n  callProc3;\nDEFAULT:\n  callDefaultProc;\n  STOP;\nENDTEST",
    ],
)
def test_test_stmt(input_str: str) -> None:
    reset_level()
    assert test_stmt.parseString(input_str, parseAll=True).as_list()


@pytest.mark.parametrize(
    "input_str",
    [
        "WHILE NOT A OR B DO\n  callProc1;\n  callProc2;\nENDWHILE",
    ],
)
def test_while_stmt(input_str: str) -> None:
    reset_level()
    assert while_stmt.parseString(input_str, parseAll=True).as_list()


@pytest.mark.parametrize(
    "input_str",
    [
        "FOR i FROM 0 TO 10 STEP 2 DO\n  callProc1;\n  callProc2;\nENDFOR",
    ],
)
def test_for_stmt(input_str: str) -> None:
    reset_level()
    assert for_stmt.parseString(input_str, parseAll=True).as_list()


@pytest.mark.parametrize(
    "input_str",
    [
        "callProc;",
        "callProc arg1, arg2, A AND B;",
        "callProc arg1, arg2, \\switch;",
        "callProc arg1, arg2, \\opt:=(1+1), \\switch;",
        "callProc arg1, name:=arg2 \\opt:=(1+1) \\switch;",
        "MoveL RelTool(target, 0, 0, -Abs(z)), v1000, fine, tool, \\WObj:=wobj0;",
    ],
)
def test_proc_call_stmt(input_str: str) -> None:
    assert proc_call_stmt.parseString(input_str, parseAll=True).as_list()


@pytest.mark.parametrize(
    "input_str",
    [
        "callFunc();",
        "callFunc(arg1, arg2, A AND B);",
        "callFunc(arg1, arg2, \\switch);",
        "callFunc(arg1, arg2, \\opt:=(1+1), \\switch);",
        "callFunc(arg1, name:=arg2 \\opt:=(1+1) \\switch);",
    ],
)
def test_func_call_stmt(input_str: str) -> None:
    assert func_call_stmt.parseString(input_str, parseAll=True).as_list()


@pytest.mark.parametrize(
    "input_str",
    [
        "CONNECT varName WITH something;",
    ],
)
def test_connect_stmt(input_str: str) -> None:
    assert connect_stmt.parseString(input_str, parseAll=True).as_list()


@pytest.mark.parametrize(
    "input_str",
    ["RETURN TRUE;", "RETURN 1+1;", "RETURN NOT (A OR B OR C);"],
)
def test_return_stmt(input_str: str) -> None:
    assert return_stmt.parseString(input_str, parseAll=True).as_list()


@pytest.mark.parametrize(
    "input_str",
    [
        "FUNC bool funcName()\nENDFUNC",
        "FUNC bool funcName(num arg1, \\num arg2, \\switch aa)\n  statement;\n  RETURN 1+1;\nENDFUNC",
    ],
)
def test_func_def(input_str: str) -> None:
    reset_level()
    assert func_def.parseString(input_str, parseAll=True).as_list()


@pytest.mark.parametrize(
    "input_str",
    [
        "PROC procName()\nENDPROC",
        "PROC procName(num arg1, \\num arg2, \\switch aa)\n  statement;\n  statement2;\nENDPROC",
        "PROC procName(num arg1, \\num arg2, \\switch aa)\n  statement;\nERROR\n  statement2;\nENDPROC",
    ],
)
def test_proc_def(input_str: str) -> None:
    reset_level()
    assert proc_def.parseString(input_str, parseAll=True).as_list()


@pytest.mark.parametrize(
    "input_str",
    [
        "TRAP trapName\nENDTRAP",
        "TRAP trapName\n  statement1;\n  statement2;\nENDTRAP",
    ],
)
def test_trap_def(input_str: str) -> None:
    reset_level()
    assert trap_def.parseString(input_str, parseAll=True).as_list()


@pytest.mark.parametrize(
    "input_str",
    [
        "TRAP trapName\nENDTRAP\nFUNC bool funcName()\nENDFUNC\nPROC procName()\nENDPROC\n",
    ],
)
def test_def_section(input_str: str) -> None:
    reset_level()
    assert def_section.parseString(input_str, parseAll=True).as_list()


def test_empty_def_section() -> None:
    reset_level()
    assert def_section.parseString("", parseAll=True).as_list() == []


@pytest.mark.parametrize(
    "input_str",
    [
        "MODULE ModuleName\nENDMODULE\n",
        "MODULE ModuleName(SYSMODULE)\n  VAR num aa;\n  FUNC bool funcName()\n    statement;\n  ENDFUNC\n  PROC procName()\n  ENDPROC\nENDMODULE",
    ],
)
def test_module(input_str: str) -> None:
    reset_level()
    assert module.parseString(input_str, parseAll=True).as_list()



from pyparsing import QuotedString, pyparsing_common

from .tokens import RapidKeyword

NUMBER = pyparsing_common.number
STRING = (QuotedString('"') | QuotedString("'")).set_name("string")
BOOL = (RapidKeyword("TRUE") | RapidKeyword("FALSE")).set_name("string")

RAPIDLITERAL = NUMBER | STRING | BOOL



from pathlib import Path

from rapidchecker.parser.grammar import module
from rapidchecker.parser.indent import reset_level


def test_simple_module_file() -> None:
    reset_level()
    file_contents = Path("./tests/rapid_samples/SimpleModule.sys").read_text()
    assert module.parseString(file_contents, parseAll=True).as_list()



from pyparsing import CaselessKeyword, ParseResults, oneOf

from rapidchecker.config import CONFIG


def case_check(s: str, loc: int, tokens: ParseResults) -> bool:
    return s[loc : loc + len(tokens[0])].isupper()


class RapidKeyword(CaselessKeyword):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if CONFIG.require_uppercase_keywords:
            self.addCondition(case_check, message="Keyword needs to be in uppercase")


# TODO: Can we turn these into enums?
MODULE = RapidKeyword("MODULE")
ENDMODULE = RapidKeyword("ENDMODULE")
VAR = RapidKeyword("VAR")
PERS = RapidKeyword("PERS")
FUNC = RapidKeyword("FUNC")
ENDFUNC = RapidKeyword("ENDFUNC")
PROC = RapidKeyword("PROC")
ENDPROC = RapidKeyword("ENDPROC")
TRAP = RapidKeyword("TRAP")
ENDTRAP = RapidKeyword("ENDTRAP")
IF = RapidKeyword("IF")
ENDIF = RapidKeyword("ENDIF")
ELSE = RapidKeyword("ELSE")
ELSEIF = RapidKeyword("ELSEIF")
THEN = RapidKeyword("THEN")
WHILE = RapidKeyword("WHILE")
ENDWHILE = RapidKeyword("ENDWHILE")
FOR = RapidKeyword("FOR")
ENDFOR = RapidKeyword("ENDFOR")
DO = RapidKeyword("DO")
FROM = RapidKeyword("FROM")
TO = RapidKeyword("TO")
STEP = RapidKeyword("STEP")
RETURN = RapidKeyword("RETURN")
CONNECT = RapidKeyword("CONNECT")
WITH = RapidKeyword("WITH")
CONST = RapidKeyword("CONST")
AND = RapidKeyword("AND")
OR = RapidKeyword("OR")
NOT = RapidKeyword("NOT")
DIV = RapidKeyword("DIV")
MOD = RapidKeyword("MOD")
XOR = RapidKeyword("XOR")
RECORD = RapidKeyword("RECORD")
ENDRECORD = RapidKeyword("ENDRECORD")
TEST = RapidKeyword("TEST")
ENDTEST = RapidKeyword("ENDTEST")
CASE = RapidKeyword("CASE")
DEFAULT = RapidKeyword("DEFAULT")
ERROR = RapidKeyword("ERROR")
INOUT = RapidKeyword("INOUT")
TASK = RapidKeyword("TASK")
LOCAL = RapidKeyword("LOCAL")

RESERVED_WORD = (
    MODULE
    | ENDMODULE
    | VAR
    | PERS
    | FUNC
    | ENDFUNC
    | PROC
    | ENDPROC
    | TRAP
    | ENDTRAP
    | IF
    | ENDIF
    | ELSE
    | ELSEIF
    | THEN
    | WHILE
    | ENDWHILE
    | FOR
    | ENDFOR
    | DO
    | FROM
    | TO
    | STEP
    | RETURN
    | CONNECT
    | WITH
    | CONST
    | AND
    | OR
    | NOT
    | DIV
    | MOD
    | XOR
    | RECORD
    | ENDRECORD
    | TEST
    | ENDTEST
    | CASE
    | DEFAULT
    | ERROR
    | INOUT
    | TASK
    | LOCAL
)

MODULE_OPTIONS = oneOf("SYSMODULE NOSTEPIN VIEWONLY READONLY")



import pyparsing as pp

from rapidchecker.config import CONFIG

indent_level = 0
INDENT_SIZE = 2
CHECK_INDENT = True


def check_indent(s: str, loc: int, _: pp.ParseResults) -> None:
    global indent_level
    if not CONFIG.indentation_check:
        return

    # Why is this necessary?
    while s[loc].strip() == "":
        loc += 1

    cur_col = pp.col(loc, s)
    expected_col = indent_level * CONFIG.indentation_size + 1

    if cur_col != expected_col:
        raise pp.ParseFatalException(
            s,
            loc,
            f"Bad indentation, (expected col {expected_col}, starts at col {cur_col})",
        )


def add_indent() -> None:
    global indent_level
    indent_level += 1


def remove_indent() -> None:
    global indent_level
    indent_level -= 1


def reset_level() -> None:
    global indent_level
    indent_level = 0


def register_indent_checks(parse_elements: list[pp.ParserElement]) -> None:
    for p in parse_elements:
        p.add_parse_action(check_indent)


INDENT = pp.Empty().set_parse_action(add_indent)
UNDENT = pp.Empty().set_parse_action(remove_indent)
INDENT_CHECKPOINT = pp.Empty().set_parse_action(check_indent)
